import os
from requests_futures.sessions import FuturesSession

import typer
from savvihub.common.store import get_token

from savvihub.savvihub_cli.config import config
from savvihub.savvihub_cli.savvihub import SavviHubClient, SavviDatasetFile
from savvihub.savvihub_cli.utils import calculate_crc32c, read_in_chunks

dataset_app = typer.Typer()


def parse_dataset_arg(dataset_arg):
    workspace, rest = dataset_arg.split("/")
    if ":" in rest:
        dataset, ref = rest.split(":")
    else:
        dataset = rest
        ref = "latest"
    return workspace, dataset, ref


@dataset_app.callback()
def main():
    return


@dataset_app.command()
def create():
    return


@dataset_app.command()
def push(
    dataset_arg: str = typer.Option(..., "-r"),
    root_path_arg: str = typer.Argument("."),
):
    root_path = os.path.abspath(root_path_arg)

    workspace, dataset, ref = parse_dataset_arg(dataset_arg)

    # download file list
    client = SavviHubClient(token=config.token)
    savvihub_files = client.dataset_file_list(workspace, dataset, ref=ref)

    # make hash map
    hash_map = dict()
    for file in savvihub_files:
        hash_map[file.path] = file.hash

    def path_and_hash(root, name):
        name = os.path.join(os.path.abspath(root), name)
        name = name[len(root_path)+1:] if name.startswith(root_path) else name
        return name, calculate_crc32c(name)

    files = []
    for root, _, files_ in os.walk(root_path):
        for name in files_:
            name, h = path_and_hash(root, name)
            if name in hash_map and hash_map[name] == h:
                continue
            files.append(name)

    typer.echo(f'Find {len(files)} files to upload.')

    with typer.progressbar(length=len(files)) as progress:
        session = FuturesSession(max_workers=config.parallel)
        client = SavviHubClient(token=config.token, session=session)

        # create files
        futures = [client.dataset_file_create(
            workspace, dataset, file_, False,
        ) for file_ in files]
        resps = [future.result() for future in futures]

        # upload files
        futures = [session.put(
            SavviDatasetFile(resp).upload_url,
            data=read_in_chunks(os.path.join(root_path, files[i])),
            headers={
                'content-type': 'application/octet-stream',
            },
            callback=lambda resp, **kwargs: resp.raise_for_status() or progress.update(1)
        ) for i, resp in enumerate(resps)]
        [future.result() for future in futures]

    typer.echo(f'Uploaded {len(files)} files in {os.path.abspath(root_path)}')


@dataset_app.command()
def pull(
    token: str = typer.Option(config.token, "-t", "--token"),
    parallel: int = typer.Option(config.parallel, "--parallel"),
    dataset_arg: str = typer.Argument(...),
    path_arg: str = typer.Argument("."),
):
    if token is None:
        token = get_token()
    if parallel:
        config.parallel = parallel

    workspace, dataset, ref = parse_dataset_arg(dataset_arg)

    client = SavviHubClient(token=token)
    files = client.dataset_file_list(workspace, dataset, ref=ref, raise_error=True)

    def mkdir(file: SavviDatasetFile, pg):
        path = os.path.join(path_arg, file.path)
        if os.path.exists(path) and os.path.isdir(path):
            pg.update(1)
            return
        os.mkdir(path)
        pg.update(1)

    def download_file(file: SavviDatasetFile, pg):
        def fn(resp, **kwargs):
            path = os.path.join(path_arg, file.path)
            print("Downloading ", path)
            os.makedirs(os.path.dirname(path), exist_ok=True)
            with open(path, 'wb') as f:
                for chunk in resp.iter_content(chunk_size=8192):
                    f.write(chunk)
            pg.update(1)
        return fn

    with typer.progressbar(length=len(files)) as progress:
        session = FuturesSession(max_workers=config.parallel)
        [mkdir(file, progress) for file in files if file.is_dir]
        futures = [session.get(
            file.download_url,
            stream=True,
            hooks={
                'response': download_file(file, progress),
            },
        ) for file in files if not file.is_dir]
        [future.result() for future in futures]

    typer.echo(f'Downloaded {len(files)} files in {os.path.abspath(path_arg)}')


if __name__ == "__main__":
    dataset_app()


import os

import typer
from requests_futures.sessions import FuturesSession
from savvihub.common.store import get_experiment, get_token

from savvihub.savvihub_cli.config import config
from savvihub.savvihub_cli.savvihub import SavviHubClient, SavviExperimentArtifact
from savvihub.savvihub_cli.utils import read_in_chunks

artifact_app = typer.Typer()


@artifact_app.callback()
def main():
    return


@artifact_app.command()
def sync(
    output_path_arg: str = typer.Argument("."),
):
    output_path = os.path.abspath(output_path_arg)

    files = []
    for root, _, files_ in os.walk(output_path):
        for name in files_:
            name = os.path.join(os.path.abspath(root), name)
            name = name[len(output_path)+1:] if name.startswith(output_path) else name
            files.append(name)

    typer.echo(f'Find {len(files)} files to upload.')

    def upload_file(path, pg):
        def fn(resp, **kwargs):
            resp.raise_for_status()
            print("Uploading ", path)
            progress.update(1)
        return fn

    with typer.progressbar(length=len(files)) as progress:
        token = get_token()
        experiment = get_experiment()

        session = FuturesSession(max_workers=config.parallel)
        client = SavviHubClient(token=token, session=session)

        # create files
        futures = [client.experiment_artifact_create(
            experiment.experiment_id,
            file_,
            hooks={
                'response': lambda resp, **kwargs: resp.raise_for_status()
            },
        ) for file_ in files]
        resps = [future.result() for future in futures]

        # upload files
        futures = [session.put(
            SavviExperimentArtifact(resp.json()).upload_url,
            data=read_in_chunks(os.path.join(output_path, files[i])),
            headers={
                'content-type': 'application/octet-stream',
            },
            hooks={
                'response': upload_file(files[i], progress),
            },
        ) for i, resp in enumerate(resps)]
        [future.result() for future in futures]

    typer.echo(f'Uploaded {len(files)} files in {os.path.abspath(output_path)}')



from jsonschema import validate, ValidationError

import yaml


class YmlLoader:
    def __init__(self, filename, schema):
        self.filename = filename
        self.schema = schema
        self.data = self._load()
        self._validate()

    def _load(self):
        try:
            with open(self.filename) as f:
                documents = yaml.full_load(f)
                if not documents:
                    return None
                data_dict = dict()
                for item, doc in documents.items():
                    data_dict[item] = doc
                return data_dict
        except EnvironmentError:
            print('Error: Config file not found')

    def _validate(self):
        try:
            validate(instance=self.data, schema=self.schema)
        except ValidationError as e:
            return e.schema["Validation Error"]

    def pop_val_by_key(self, key):
        return self.data.pop(key, None)

    def _is_duplicated(self, key, value):
        if value in self.data[key]:
            return True
        else:
            return False

    def write(self, dict_data):
        try:
            with open(self.filename, 'w') as f:
                yaml.safe_dump(dict_data, f, default_flow_style=False)
        except EnvironmentError:
            print('Error: Config file not found')

    def update_and_write(self, dict_data):
        for k, v in dict_data.items():
            if k in self.data.keys():
                if type(self.data[k]) == list:
                    if self._is_duplicated(k, v):
                        raise Exception('Duplicated value!')
                    self.data[k].append(v)
                else:
                    self.data[k] = v
            else:
                properties = self.schema.get('oneOf')[1].get('properties')
                if k in properties.keys():
                    if properties.get(k).get('type') == 'array':
                        self.data[k] = [v]
                    else:
                        self.data[k] = v
                else:
                    raise Exception('Wrong key: ', k)
        self.write(self.data)

    def is_ready_to_run(self):
        if not self.data.get('data_mount_infos', None):
            return False
        else:
            return True


class ExperimentYmlLoader(YmlLoader):
    schema = {
        "oneOf": [
            {"type": "null"},
            {
                "type": "object",
                "properties": {
                    "workspace": {"type": "string"},
                    "project": {"type": "string"},
                    "image_id": {"type": "string"},
                    "resource_spec_id": {"type": "string"},
                    "data_mount_infos": {
                        "type": "array",
                        "properties": {
                            "id": {"type": "string"},
                            "ref": {"type": "string"},
                            "mount_path": {"type": "string"}
                        },
                        "required": ["id", "ref", "mount_path"]
                    },
                    "start_command": {"type": "string"}
                },
                "required": ["workspace", "project"]
            }
        ]
    }

    def __init__(self, filename):
        super().__init__(filename, self.schema)



import collections
import numbers

import six

from savvihub.savvihub_cli.config import config as global_config
from savvihub import history


class Experiment:
    def __init__(self, experiment_id):
        self.experiment_id = experiment_id
        self._history = None

    @classmethod
    def from_given_or_global_config(cls, config):
        if not config:
            config = global_config
        experiment = cls(config.experiment_id)
        return experiment

    @property
    def history(self):
        if not self._history:
            self._history = history.History(self)
        return self._history

    def log(self, step, row=None):
        if not row:
            row = dict()

        if not isinstance(row, collections.Mapping):
            raise ValueError(".log() takes a dictionary as a parameter")

        if any(not isinstance(key, six.string_types) for key in row.keys()):
            raise ValueError("The key of dictionary in .log() parameter must be str")

        for k in row.keys():
            if not k:
                raise ValueError("Logging empty key is not supported")

        if not isinstance(step, numbers.Number):
            raise ValueError(f"Step must be a number, not {type(step)}")
        if step < 0:
            raise ValueError(f"Step must be a positive integer, not {step}")
        if not isinstance(type(step), int):
            step = int(round(step))

        self.history.update(step, row)



import requests
import typing

from savvihub.savvihub_cli.constants import API_HOST
from savvihub.savvihub_cli.utils import AnnotatedObject


class SavviDatasetFile(AnnotatedObject):
    path: str
    is_dir: bool

    size: int
    hash: str

    download_url: str
    upload_url: str


class SavviExperimentArtifact(AnnotatedObject):
    path: str
    is_dir: bool

    size: int
    hash: str

    download_url: str
    upload_url: str


class SavviKernelImage(AnnotatedObject):
    id: int
    image_url: str
    name: str


class SavviKernelResource(AnnotatedObject):
    id: int
    name: str
    cpu_limit: float
    mem_limit: str


class SavviDataset(AnnotatedObject):
    id: int
    name: str


class SavviListResponse(AnnotatedObject):
    results: typing.List


class PaginatedMixin(AnnotatedObject):
    total: int
    startCursor: typing.Optional[str]
    endCursor: typing.Optional[str]
    results: typing.List


class SavviHubPaginatedDatasetFilesResponse(PaginatedMixin, AnnotatedObject):
    results: typing.List[SavviDatasetFile]


class SavviHubClient:
    def __init__(self, *, session=requests.Session(), token=None, url=API_HOST, content_type='application/json'):
        self.session = session
        self.url = url
        self.token = token
        
        session.headers = {'content-type': content_type}
        if token:
            session.headers['authorization'] = 'Token %s' % token

    def get(self, url, params=None, raise_error=False, **kwargs):
        r = self.session.get(f'{self.url}{url}', params=params, **kwargs)
        if raise_error:
            r.raise_for_status()
        return r

    def get_all(self, url, params=None, raise_error=False, **kwargs):
        raw_resp = self.get(url, params=params, raise_error=raise_error, **kwargs)
        resp = PaginatedMixin(raw_resp.json())
        results = []

        fetched_items = 0
        while True:
            fetched_items += len(resp.results)
            results.extend(resp.results)
            if fetched_items >= resp.total:
                break
            raw_resp = self.get(url, params={**params, 'after': resp.endCursor}, raise_error=raise_error, **kwargs)
            resp = PaginatedMixin(raw_resp.json())
        return results

    def get_all_without_pagination(self, url, params=None, raise_error=False, **kwargs):
        raw_resp = self.get(url, params=params, raise_error=raise_error, **kwargs)
        resp = SavviListResponse(raw_resp.json())
        return resp.results

    def post(self, url, data, raise_error=False, **kwargs):
        r = self.session.post(f'{self.url}{url}', json=data, **kwargs)
        if raise_error:
            r.raise_for_status()
        return r

    def delete(self, url, raise_error=False, **kwargs):
        r = self.session.delete(f'{self.url}{url}', **kwargs)
        if raise_error:
            r.raise_for_status()
        return r

    def patch(self, url, data, raise_error=False, **kwargs):
        r = self.session.patch(f'{self.url}{url}', json=data, **kwargs)
        if raise_error:
            r.raise_for_status()
        return r

    def dataset_file_list(self, workspace, dataset, *, ref='latest', **kwargs) -> typing.List[SavviDatasetFile]:
        results = self.get_all(f'/v1/api/workspaces/{workspace}/datasets/{dataset}/files/', params={'ref': ref}, **kwargs)
        return [SavviDatasetFile(x) for x in results]

    def dataset_file_create(self, workspace, dataset, path, is_dir, **kwargs):
        return self.post(f'/v1/api/workspaces/{workspace}/datasets/{dataset}/files/', {
            'path': path,
            'is_dir': is_dir
        }, **kwargs)

    def experiment_read(self, **kwargs):
        return self.get(f'/v1/api/experiments/', **kwargs)

    def experiment_list(self, workspace, project, **kwargs):
        return self.get(f'/v1/api/workspaces/{workspace}/projects/{project}/experiments/?orderby.direction=DESC', **kwargs)

    def experiment_log(self, workspace, project, experiment_number, **kwargs):
        return self.get(f'/v1/api/workspaces/{workspace}/projects/{project}/experiments/{experiment_number}/log/', **kwargs)

    def experiment_create(self, workspace, project, image_id, resource_spec_id, branch, dataset_mount_infos, start_command, **kwargs):
        return self.post(f'/v1/api/workspaces/{workspace}/projects/{project}/experiments/', {
            'image_id': image_id,
            'resource_spec_id': resource_spec_id,
            'branch': branch,
            'dataset_mount_infos': dataset_mount_infos,
            'start_command': start_command,
        }, **kwargs)

    def experiment_progress_update(self, experiment_id, row, **kwargs):
        return self.post(f'/v1/api/experiments/{experiment_id}/progress/', {
            'metrics': [row],
        }, **kwargs)

    def experiment_artifact_create(self, experiment_id, path, **kwargs):
        return self.post(f'/v1/api/experiments/{experiment_id}/artifacts/', {
            'path': path,
            'is_dir': False,
        }, **kwargs)

    def kernel_image_list(self, workspace):
        results = self.get_all_without_pagination(f'/v1/api/workspaces/{workspace}/kernels/images/')
        return [SavviKernelImage(x) for x in results]

    def kernel_resource_list(self, workspace):
        results = self.get_all_without_pagination(f'/v1/api/workspaces/{workspace}/kernels/resource_specs')
        return [SavviKernelResource(x) for x in results]

    def public_dataset_list(self):
        results = self.get_all(f'/v1/api/datasets/public/')
        return [SavviDataset(x) for x in results]

    def dataset_list(self, workspace):
        results = self.get_all(f'/v1/api/workspaces/{workspace}/datasets/')
        if not results:
            return []
        return [SavviDataset(x) for x in results]



from savvihub.common.store import get_experiment
from savvihub.savvihub_cli.config import config
from savvihub.savvihub_cli.constants import DEFAULT_CONFIG_PATH
from savvihub.savvihub_cli.savvihub import SavviHubClient

from savvihub.experiment import Experiment
from savvihub.savvihub_cli.utils import get_token_from_config

experiment = None


def log(step, row=None):
    """
    step: a step for each iteration (required)
    row: a dictionary to log
    """
    experiment = get_experiment()
    experiment.log(step, row)
